﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;

namespace ConsoleApplication84
{
    class Program
    {
        const int sampleRate = 192000;

        Program()
        {
            double baudRate = 115200;

            short[] srcSignal = LoadWav("signal.wav");

            var bits = new List<bool>();
            var bytes = new List<byte>();

            int detect_level = 2480;

            int chunkStart = 0;
            int lockIndex = 0;
            short lockValue = 0;

            int bitsReceived = 0;
            int bitErrors = 0;

            var hpf = new double[] { -2.03324, 3.03324 };

            var recoveredSignal = Convolution(hpf, srcSignal);

            var deltaSignal = new short[recoveredSignal.Length - 1];
            for (int i = 0; i < recoveredSignal.Length - 1; i++)
                deltaSignal[i] = (short)(recoveredSignal[i + 1] - recoveredSignal[i]);

            int oversampleFactor = 4;
            deltaSignal = Interpolation(deltaSignal, oversampleFactor);

            int resampleFilterSize = oversampleFactor * 8 + 1;
            deltaSignal = Convolution(MakeLowPassFilter(
                sampleRate / 2, resampleFilterSize, sampleRate * oversampleFactor), deltaSignal);

            for (int i = 0; i < deltaSignal.Length; i++)
            {
                short deltaSample = deltaSignal[i];
                short deltaAbsSample = Math.Abs(deltaSample);
                if ((lockIndex == 0 && deltaAbsSample > detect_level) ||
                    (lockIndex != 0 && deltaAbsSample >= lockValue))
                {
                    lockIndex = i;
                    lockValue = deltaAbsSample;
                }
                else if (lockIndex != 0 && deltaAbsSample < lockValue)
                {
                    int bitCount = (int)Math.Round(baudRate * (i - chunkStart) / sampleRate / oversampleFactor);
                    if (bitCount > 9)
                        bitCount = 9;
                    bool bitsValue = deltaSample > 0;
                    for (int x = 0; x < bitCount; x++)
                        bits.Add(bitsValue);
                    bitsReceived += bitCount;
                    while (bits.Count != 0 && bits[0])
                        bits.RemoveAt(0);

                    if (bits.Count > 9)
                    {
                        int value = 0;
                        for (int x = 0; x < 8; x++)
                        {
                            value >>= 1;
                            value |= bits[x + 1] ? 0x80 : 0;
                        }
                        if (!bits[9])
                            bitErrors++;
                        bytes.Add((byte)value);
                        bits.RemoveRange(0, 10);
                    }
                    chunkStart = lockIndex;
                    lockIndex = 0;
                }
            }

            Console.WriteLine($"Bits received: {bitsReceived}");
            Console.WriteLine($"Bit errors   : {bitErrors}");
            File.WriteAllBytes("decoded.txt", bytes.ToArray());
        }

        static void Main(string[] args)
        {
            new Program();
        }

        double[] MakeLowPassFilter(double CutoffFreq, int SampleCount, int sampleRate)
        {
            double W = 2 * Math.PI * (CutoffFreq / sampleRate);
            double A = W / Math.PI;

            int N = SampleCount / 2;

            double[] IR = new double[SampleCount];

            for (int i = 0; i < N + 1; i++)
            {
                if (i == 0)
                    IR[N] = A;
                else
                {
                    double V = A * (Math.Sin(i * W) / (i * W));
                    IR[N + i] = V;
                    IR[N - i] = V;
                }
            }
            return IR;
        }

        double Lerp(double a, double b, double t)
        {
            return a + (b - a) * t;
        }

        short[] Interpolation(short[] signal, int factor)
        {
            if (factor <= 0)
                throw new Exception();
            if (factor == 1)
                return signal;
            if (signal.Length == 0)
                return signal;

            short[] result = new short[(signal.Length - 1) * factor + 1];
            for (int i = 0; i < signal.Length - 1; i++)
            {
                for (int j = 0; j < factor; j++)
                {
                    double t = (double)j / factor;
                    result[i * factor + j] = (short)Lerp(signal[i], signal[i + 1], t);
                }
            }
            result[(signal.Length - 1) * factor] = signal[signal.Length - 1];
            return result;
        }

        short[] Convolution(double[] W1, short[] W2)
        {
            if (W2.Length <= W1.Length)
                return new short[0];

            short[] R = new short[W2.Length - W1.Length];

            int blockSize = 512;
            int blockCount = (R.Length + blockSize - 1) / blockSize;

            Parallel.For(0, blockCount, i =>
            {
                int startSample = i * blockSize;
                int xendSample = blockSize;
                if (i == blockCount - 1)
                    xendSample = R.Length - blockSize * i;
                xendSample += startSample;
                for (int j = startSample; j < xendSample; j++)
                {
                    double Sum = 0;
                    for (int k = 0; k < W1.Length; k++)
                        Sum += W1[k] * W2[j + k];
                    R[j] = (short)Sum;
                }
            });
            return R;
        }

        short[] LoadWav(string Path)
        {
            FileStream fs = File.Open(Path, FileMode.Open);
            var br = new BinaryReader(fs);
            br.ReadBytes(12);
            int sampleCount = 0;
            for (;;)
            {
                uint cid = br.ReadUInt32();
                int csize = br.ReadInt32();
                if (cid == 0x61746164)
                {
                    sampleCount = csize / 2;
                    break;
                }
                br.ReadBytes(csize);
            }

            short[] signal = new short[sampleCount];
            for (int i = 0; i < sampleCount; i++)
                signal[i] = br.ReadInt16();
            fs.Close();
            return signal;
        }
    }
}
